import importlib
import json
import logging
from contextlib import contextmanager
from pathlib import Path
from typing import Callable, Iterator, Optional, Tuple

import click
from sagemaker.workflow.pipeline import Pipeline as SageMakerPipeline

from kedro_sagemaker.constants import MLFLOW_TAG_EXECUTION_ARN
from kedro_sagemaker.generator import KedroSageMakerGenerator
from kedro_sagemaker.utils import (
    CliContext,
    KedroContextManager,
    docker_build,
    docker_push,
)

logger = logging.getLogger()


def parse_extra_params(params, silent=False):
    if params and (parameters := json.loads(params.strip("'"))):
        if not silent:
            click.echo(
                f"Running with extra parameters:\n{json.dumps(parameters, indent=4)}"
            )
    else:
        parameters = None
    return parameters


@contextmanager
def get_context_and_pipeline(
    ctx: CliContext,
    docker_image: Optional[str],
    pipeline: str,
    params: str,
    is_local: bool,
    execution_role_arn: Optional[str] = None,
) -> Iterator[Tuple[KedroContextManager, SageMakerPipeline]]:
    with KedroContextManager(
        ctx.metadata.package_name, ctx.env, parse_extra_params(params, True)
    ) as mgr:
        generator = KedroSageMakerGenerator(
            pipeline,
            mgr.context,
            mgr.plugin_config,
            docker_image,
            is_local,
            execution_role_arn,
        )
        sm_pipeline = generator.generate()
        yield mgr, sm_pipeline


def docker_autobuild(auto_build, click_context, image, mgr, yes):
    if auto_build:
        if (splits := image.split(":"))[-1] != "latest" and len(splits) > 1:
            click.echo(
                click.style(
                    f"This operation will overwrite the target image with {splits[-1]} tag at remote location.",
                    fg="yellow",
                )
            )

        if not yes and not click.confirm("Continue?", default=True):
            click_context.exit(1)

        if (rv := docker_build(str(mgr.context.project_path), image)) != 0:
            click_context.exit(rv)
        if (rv := docker_push(image)) != 0:
            click_context.exit(rv)
    else:
        click.echo(
            click.style(
                "Make sure that you've built and pushed your image to run the latest version remotely.\
 Consider using '--auto-build' parameter.",
                fg="yellow",
            )
        )


def write_file_and_confirm_overwrite(
    filepath: Path,
    auto_overwrite: bool,
    contents: str,
    on_denied_overwrite: Callable[[Path], None] = None,
):
    write = True
    if filepath.exists():
        write = False
        if auto_overwrite or click.confirm(
            f"{filepath} exists, overwrite?", default=True
        ):
            write = True
    if write:
        filepath.write_text(contents)
    elif on_denied_overwrite:
        on_denied_overwrite(filepath)


def lookup_mlflow_run_id(context, sagemaker_execution_arn: str):
    import mlflow
    from kedro_mlflow.config.kedro_mlflow_config import KedroMlflowConfig

    mlflow_conf: KedroMlflowConfig = context.mlflow
    mlflow_runs = mlflow.search_runs(
        experiment_names=[mlflow_conf.tracking.experiment.name],
        filter_string=f'tags.`{MLFLOW_TAG_EXECUTION_ARN}` = "{sagemaker_execution_arn}"',
        max_results=1,
        output_format="list",
    )
    importlib.reload(mlflow.tracking.request_header.registry)

    if len(mlflow_runs) == 0:
        logger.warning(
            "Unable to find parent mlflow run id for the current execution (%s)",
            sagemaker_execution_arn,
        )
        return mlflow.tracking._RUN_ID_ENV_VAR, None

    return mlflow.tracking._RUN_ID_ENV_VAR, mlflow_runs[0].info.run_id



import logging
from functools import wraps

from kedro_sagemaker.config import SageMakerMetricsTrackingConfig
from kedro_sagemaker.constants import KEDRO_SAGEMAKER_METRICS

logger = logging.getLogger(__name__)


def sagemaker_metrics(metrics: dict):
    def _decorator(func):
        config = SageMakerMetricsTrackingConfig(metrics)
        setattr(
            func,
            KEDRO_SAGEMAKER_METRICS,
            config,
        )

        @wraps(func)
        def wrapper(*args, **kws):
            # for later use, maybe we will actually need to plug-in custom actions
            return func(*args, **kws)

        return wrapper

    return _decorator



import logging
import os
from typing import Any, Dict

from kedro.io import AbstractDataSet, DataCatalog
from kedro.pipeline import Pipeline
from kedro.runner import SequentialRunner
from pluggy import PluginManager
from pydantic import BaseModel

from kedro_sagemaker.constants import (
    KEDRO_SAGEMAKER_EXECUTION_ARN,
    KEDRO_SAGEMAKER_RUNNER_CONFIG,
)
from kedro_sagemaker.datasets import (
    CloudpickleDataset,
    DistributedCloudpickleDataset,
)
from kedro_sagemaker.utils import is_distributed_environment

logger = logging.getLogger(__name__)


class KedroSageMakerRunnerConfig(BaseModel):
    bucket: str


class SageMakerPipelinesRunner(SequentialRunner):
    @classmethod
    def runner_name(cls):
        return f"{cls.__module__}.{cls.__qualname__}"

    def __init__(self, is_async: bool = False):
        super().__init__(is_async)
        self.runner_config_raw = os.environ.get(KEDRO_SAGEMAKER_RUNNER_CONFIG)
        self.runner_config = KedroSageMakerRunnerConfig.parse_raw(
            self.runner_config_raw
        )
        self.run_id = os.getenv(KEDRO_SAGEMAKER_EXECUTION_ARN, "local").split(":")[-1]

    def run(
        self,
        pipeline: Pipeline,
        catalog: DataCatalog,
        hook_manager: PluginManager = None,
        session_id: str = None,
    ) -> Dict[str, Any]:
        unsatisfied = pipeline.inputs() - set(catalog.list())
        for ds_name in unsatisfied:
            catalog = catalog.shallow_copy()
            catalog.add(ds_name, self.create_default_data_set(ds_name))

        return super().run(pipeline, catalog, hook_manager, session_id)

    def create_default_data_set(self, ds_name: str) -> AbstractDataSet:
        # TODO: handle credentials better (probably with built-in Kedro credentials
        #  via ConfigLoader (but it's not available here...)
        dataset_cls = CloudpickleDataset
        if is_distributed_environment():
            logger.info("Using distributed dataset class as a default")
            dataset_cls = DistributedCloudpickleDataset

        return dataset_cls(
            bucket=self.runner_config.bucket,
            dataset_name=ds_name,
            run_id=self.run_id,
        )



import json
import logging
import os
import subprocess
import sys
from collections import defaultdict
from dataclasses import dataclass
from functools import cached_property
from typing import Any, List, MutableMapping, Optional, Union

from kedro.framework.session import KedroSession

from kedro_sagemaker.config import KedroSageMakerPluginConfig
from kedro_sagemaker.constants import (
    KEDRO_SAGEMAKER_PARAM_KEY_PREFIX,
    KEDRO_SAGEMAKER_PARAM_VALUE_PREFIX,
)

logger = logging.getLogger()


@dataclass
class CliContext:
    env: str
    metadata: Any


def is_distributed_master_node() -> bool:
    is_rank_0 = True
    try:
        if "TF_CONFIG" in os.environ:
            # TensorFlow
            tf_config = json.loads(os.environ["TF_CONFIG"])
            worker_type = tf_config["task"]["type"].lower()
            is_rank_0 = (worker_type == "chief" or worker_type == "master") or (
                worker_type == "worker" and tf_config["task"]["index"] == 0
            )
        else:
            # MPI + PyTorch
            for e in ("OMPI_COMM_WORLD_RANK", "RANK"):
                if e in os.environ:
                    is_rank_0 = int(os.environ[e]) == 0
                    break
    except:  # noqa
        logger.error(
            "Could not parse environment variables related to distributed computing. "
            "Set appropriate values for one of: RANK, OMPI_COMM_WORLD_RANK or TF_CONFIG",
            exc_info=True,
        )
        logger.warning("Assuming this node is not a master node, due to error.")
        return False
    return is_rank_0


def is_distributed_environment() -> bool:
    return any(e in os.environ for e in ("OMPI_COMM_WORLD_RANK", "RANK", "TF_CONFIG"))


class KedroContextManager:
    def __init__(
        self, package_name: str, env: str, extra_params: Optional[dict] = None
    ):
        self.extra_params = extra_params
        self.env = env
        self.package_name = package_name
        self.session: Optional[KedroSession] = None

    @cached_property
    def plugin_config(self) -> KedroSageMakerPluginConfig:
        return KedroSageMakerPluginConfig.parse_obj(
            self.context.config_loader.get("sagemaker*")
        )

    @cached_property
    def context(self):
        assert self.session is not None, "Session not initialized yet"
        return self.session.load_context()

    def __enter__(self):
        self.session = KedroSession.create(
            self.package_name, env=self.env, extra_params=self.extra_params
        )
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.session.__exit__(exc_type, exc_val, exc_tb)


def flatten_dict(d: MutableMapping, parent_key: str = "", sep: str = ".") -> dict:
    items = []
    for k, v in d.items():
        new_key = parent_key + sep + k if parent_key else k
        if isinstance(v, MutableMapping):
            items.extend(flatten_dict(v, new_key, sep=sep).items())
        else:
            items.append((new_key, v))
    return dict(items)


def parse_flat_parameters(
    param_source: Union[dict, MutableMapping],
    key_prefix: str = KEDRO_SAGEMAKER_PARAM_KEY_PREFIX,
    value_prefix: str = KEDRO_SAGEMAKER_PARAM_VALUE_PREFIX,
) -> dict:
    """
    This function takes `param_source` (most likely os.environ)
    searches for the keys starting with `key_prefix`/`value_prefix`
    and parses them from the form:
    "key_prefix_0": "name.of.the.param"
    "value_prefix_0": "666"
    into dictionary: {"name": {"of": {"the": {"param": 666}}}}

    :param param_source:
    :param key_prefix:
    :param value_prefix:
    :return:
    """

    def sorted_keys_with_prefix(prefix) -> List[str]:
        return sorted(
            (x for x in param_source.keys() if x.startswith(prefix)),
            key=lambda x: int(x.replace(prefix, "")),
        )

    def nested_defaultdict():
        return defaultdict(nested_defaultdict)

    params = nested_defaultdict()

    keys = sorted_keys_with_prefix(key_prefix)
    values = sorted_keys_with_prefix(value_prefix)
    assert len(keys) == len(values), "Invalid parameter/key pairs"

    for k, v in zip(keys, values):
        p = params
        key = param_source[k]
        value = param_source[v]
        while "." in key:
            level, reminder = key.split(".", maxsplit=1)
            p = p[level]
            key = reminder
        else:
            try:
                p[key] = json.loads(value)
            except json.JSONDecodeError:
                p[key] = value
    return params


def docker_build(path: str, image: str) -> int:
    rv = subprocess.run(
        [
            "docker",
            "build",
            path,
            "-t",
            image,
        ],
        stdout=sys.stdout,
        stderr=subprocess.STDOUT,
    ).returncode
    if rv:
        logger.error("Docker build has failed.")
    return rv


def docker_push(image: str) -> int:
    rv = subprocess.run(
        ["docker", "push", image], stdout=sys.stdout, stderr=subprocess.STDOUT
    ).returncode
    if rv:
        logger.error("Docker push has failed.")
    return rv


def is_mlflow_enabled() -> bool:
    try:
        import kedro_mlflow  # NOQA
        import mlflow  # NOQA

        return True
    except ImportError:
        return False



import logging
from copy import deepcopy
from functools import lru_cache
from pathlib import Path
from sys import version_info
from typing import Any, Dict, Union

import backoff
import cloudpickle
import fsspec
import zstandard as zstd
from kedro.io import AbstractDataSet, DataSetError
from tarsafe import TarSafe

from kedro_sagemaker.constants import KEDRO_SAGEMAKER_S3_TEMP_DIR_NAME
from kedro_sagemaker.utils import is_distributed_master_node

logger = logging.getLogger(__name__)


class SageMakerModelDataset(AbstractDataSet):
    STORE_METHODS = ("pickle", "blob", "cloudpickle")
    SAGEMAKER_DEFAULT_PATH = "/opt/ml"
    MODEL_TAR_GZ = "model.tar.gz"

    def __init__(
        self,
        store_method: str = "pickle",
        sagemaker_path: str = "/opt/ml/model",
        load_kwargs: dict = None,
        save_kwargs: dict = None,
        model_registry_kwargs: dict = None,
    ):
        super().__init__()
        assert (
            store_method in self.STORE_METHODS
        ), f"Unsupported store method {store_method}, possible: {','.join(self.STORE_METHODS)}"
        assert load_kwargs is None or isinstance(load_kwargs, dict)
        assert save_kwargs is None or isinstance(save_kwargs, dict)
        self.store_method = store_method
        self.load_kwargs = load_kwargs or {}
        self.save_kwargs = save_kwargs or {}
        self.model_registry_kwargs = model_registry_kwargs or {}
        self.sagemaker_path = sagemaker_path

    def _load(self) -> Union[Any, dict]:
        load_kwargs = deepcopy(self.load_kwargs)
        search_path = load_kwargs.pop("path", None) or self.SAGEMAKER_DEFAULT_PATH
        for p in Path(search_path).rglob(self.MODEL_TAR_GZ):
            if p.is_file():
                with TarSafe.open(p, "r:gz") as tf:
                    tf.extractall(self.sagemaker_path)
                    break
        else:
            raise DataSetError(f"No model.tar.gz found in {search_path}")

        to_load = {}
        for p in Path(self.sagemaker_path).iterdir():
            if (
                p.is_file()
                and not p.is_symlink()
                and not p.name.startswith(".")
                and p.name != self.MODEL_TAR_GZ
            ):
                logger.info(f"File to load: {p.absolute()}")
                to_load[p.name] = p

        def _pickle(fo):
            import pickle

            return pickle.load(fo, **(load_kwargs or {}))

        def _cloudpickle(fo):
            import cloudpickle

            return cloudpickle.load(fo, **(load_kwargs or {}))

        methods = {
            "blob": lambda fo: fo.read(),
            "pickle": _pickle,
            "cloudpickle": _cloudpickle,
        }

        results = {}
        for name, path in to_load.items():
            with path.open("rb") as f:
                results[name] = methods[self.store_method](f)

        if len(results) == 1:
            return next(iter(results.values()))
        else:
            return results

    def _save(self, data: Union[Any, dict]) -> None:
        save_kwargs = deepcopy(self.save_kwargs)
        to_save = (
            data
            if isinstance(data, dict)
            else {save_kwargs.pop("model_file_name", "model.pkl"): data}
        )

        def _pickle(f, obj):
            import pickle

            return pickle.dump(obj, f, **(save_kwargs or {}))

        def _cloudpickle(f, obj):
            import cloudpickle

            return cloudpickle.dump(obj, f, **(save_kwargs or {}))

        methods = {
            "blob": lambda f, obj: f.write(obj),
            "pickle": _pickle,
            "cloudpickle": _cloudpickle,
        }

        for name, data in to_save.items():
            with (Path(self.sagemaker_path) / name).open("wb") as f:
                methods[self.store_method](f, data)

    def _describe(self) -> Dict[str, Any]:
        return {}


class CloudpickleDataset(AbstractDataSet):
    def __init__(
        self,
        bucket,
        dataset_name,
        run_id,
    ):
        self.bucket = bucket
        self.run_id = run_id
        self.dataset_name = dataset_name
        self.pickle_protocol = None if version_info[:2] > (3, 8) else 4

    @lru_cache()
    def _get_target_path(self):
        # This fails with:
        # 2022-11-29T13:19:22.871+01:00	DataSetError: Failed while saving data to data set
        #
        # 2022-11-29T13:19:22.871+01:00	CloudpickleDataset(dataset_name=data_processing.preprocessed_shuttles, info=for
        #
        # 2022-11-29T13:19:22.871+01:00	use only within SageMaker Pipelines,
        #
        # 2022-11-29T13:19:22.871+01:00	path=/opt/ml/processing/data_processing.preprocessed_shuttles/data.bin).
        #
        # 2022-11-29T13:19:22.871+01:00	[Errno 13] Permission denied:
        #
        # 2022-11-29T13:19:22.871+01:00	'/opt/ml/processing/data_processing.preprocessed_shuttles/data.bin'
        # p = Path(
        #     f"/opt/ml/processing/{self.dataset_name}"  # TODO - move to config
        # )
        # p.mkdir(parents=True, exist_ok=True)
        # return str(p / "data.bin")
        return f"s3://{self.bucket}/{KEDRO_SAGEMAKER_S3_TEMP_DIR_NAME}/{self.run_id}/{self.dataset_name}.bin"

    @lru_cache()
    def _get_storage_options(self):
        return {}

    def _load(self):
        with fsspec.open(
            self._get_target_path(), "rb", **self._get_storage_options()
        ) as f:
            with zstd.open(f, "rb") as stream:
                return cloudpickle.load(stream)

    def _save(self, data: Any) -> None:
        with fsspec.open(
            self._get_target_path(), "wb", **self._get_storage_options()
        ) as f:
            with zstd.open(f, "wb") as stream:
                cloudpickle.dump(data, stream, protocol=self.pickle_protocol)

    def _describe(self) -> Dict[str, Any]:
        return {
            "info": "for use only within SageMaker Pipelines",
            "dataset_name": self.dataset_name,
            "path": self._get_target_path(),
        }


class DistributedCloudpickleDataset(CloudpickleDataset):
    @backoff.on_exception(
        backoff.fibo,
        Exception,
        max_time=lambda: 300,
        raise_on_giveup=False,
    )
    def _load(self):
        return super()._load()

    def _save(self, data: Any) -> None:
        if is_distributed_master_node():
            super()._save(data)
        else:
            logger.warning(
                f"DataSet {self.dataset_name} will not be saved on a distributed node"
            )



KEDRO_SAGEMAKER = "KEDRO_SAGEMAKER"
KEDRO_SAGEMAKER_METRICS = f"{KEDRO_SAGEMAKER}_METRICS"
KEDRO_SAGEMAKER_RUNNER_CONFIG = f"{KEDRO_SAGEMAKER}_RUNNER_CONFIG"
KEDRO_SAGEMAKER_ARGS = f"{KEDRO_SAGEMAKER}_ARGS"
KEDRO_SAGEMAKER_DEBUG = f"{KEDRO_SAGEMAKER}_DEBUG"
KEDRO_SAGEMAKER_WORKING_DIRECTORY = f"{KEDRO_SAGEMAKER}_WD"
KEDRO_SAGEMAKER_PARAMETERS = f"{KEDRO_SAGEMAKER}_PARAMETERS"
KEDRO_SAGEMAKER_EXECUTION_ARN = f"{KEDRO_SAGEMAKER}_EXECUTION_ARN"
KEDRO_SAGEMAKER_PARAM_KEY_PREFIX = f"{KEDRO_SAGEMAKER}_PARAM_KEY_"
KEDRO_SAGEMAKER_PARAM_VALUE_PREFIX = f"{KEDRO_SAGEMAKER}_PARAM_VALUE_"
KEDRO_SAGEMAKER_S3_TEMP_DIR_NAME = "kedro-sagemaker-tmp"
MLFLOW_TAG_EXECUTION_ARN = "sagemaker_execution_arn"



import json
import os
import shlex
import subprocess
import sys
from pathlib import Path
from typing import Optional

import click
from kedro.framework.startup import ProjectMetadata

from kedro_sagemaker.cli_functions import (
    docker_autobuild,
    get_context_and_pipeline,
    lookup_mlflow_run_id,
    parse_extra_params,
    write_file_and_confirm_overwrite,
)
from kedro_sagemaker.client import SageMakerClient
from kedro_sagemaker.config import CONFIG_TEMPLATE_YAML
from kedro_sagemaker.constants import (
    KEDRO_SAGEMAKER_ARGS,
    KEDRO_SAGEMAKER_DEBUG,
    KEDRO_SAGEMAKER_EXECUTION_ARN,
    KEDRO_SAGEMAKER_S3_TEMP_DIR_NAME,
    KEDRO_SAGEMAKER_WORKING_DIRECTORY,
    MLFLOW_TAG_EXECUTION_ARN,
)
from kedro_sagemaker.docker import DOCKERFILE_TEMPLATE, DOCKERIGNORE_TEMPLATE
from kedro_sagemaker.runner import SageMakerPipelinesRunner
from kedro_sagemaker.utils import (
    CliContext,
    KedroContextManager,
    is_mlflow_enabled,
    parse_flat_parameters,
)


@click.group("SageMaker")
def commands():
    """Kedro plugin adding support for SageMaker Pipelines"""
    pass


@commands.group(
    name="sagemaker", context_settings=dict(help_option_names=["-h", "--help"])
)
@click.option(
    "-e",
    "--env",
    "env",
    type=str,
    default=lambda: os.environ.get("KEDRO_ENV", "local"),
    help="Environment to use.",
)
@click.pass_obj
@click.pass_context
def sagemaker_group(ctx, metadata: ProjectMetadata, env):
    click.echo(metadata)
    ctx.obj = CliContext(env, metadata)


@sagemaker_group.command()
@click.argument("bucket")
@click.argument("execution_role")
@click.argument("docker_image")
@click.option(
    "--yes",
    "-y",
    is_flag=True,
    default=False,
    type=bool,
    help="Auto answer yes confirm prompts",
)
@click.pass_obj
def init(ctx: CliContext, bucket, execution_role, docker_image, yes: bool):
    """
    Creates basic configuration for Kedro SageMaker plugin
    """
    cwd = Path.cwd()
    target_path = cwd.joinpath("conf/base/sagemaker.yml")
    cfg = CONFIG_TEMPLATE_YAML.format(
        **{
            "bucket": bucket,
            "execution_role": execution_role,
            "docker_image": docker_image,
        }
    )
    target_path.write_text(cfg)

    click.echo(f"Configuration generated in {target_path}")

    def on_denied_overwrite(filepath: Path):
        click.echo(
            click.style(
                f"Kedro SageMaker recommends to use auto-generated {filepath.name} "
                "to ensure SageMaker-compatible docker images",
                fg="yellow",
            )
        )

    write_file_and_confirm_overwrite(
        cwd / "Dockerfile", yes, DOCKERFILE_TEMPLATE, on_denied_overwrite
    )
    write_file_and_confirm_overwrite(
        cwd / ".dockerignore", yes, DOCKERIGNORE_TEMPLATE, on_denied_overwrite
    )

    click.echo(
        click.style(
            f"It's recommended to perform S3 Lifecycle configuration for bucket {bucket} "
            f"to avoid costs of long-term storage of the temporary data."
            f"\nTemporary data will be stored under s3://{bucket}/{KEDRO_SAGEMAKER_S3_TEMP_DIR_NAME} path"  # noqa
            f"\nSee https://docs.aws.amazon.com/AmazonS3/latest/userguide/object-lifecycle-mgmt.html",  # noqa
            fg="green",
        )
    )


@sagemaker_group.command()
@click.option(
    "-r",
    "--execution-role",
    type=str,
    help="SageMaker execution role",
)
@click.option(
    "-i",
    "--image",
    type=str,
    help="Docker image to use for pipeline execution.",
)
@click.option(
    "-p",
    "--pipeline",
    "pipeline",
    type=str,
    help="Name of pipeline to run",
    default="__default__",
)
@click.option(
    "--params",
    "params",
    type=str,
    help="Parameters override in form of JSON string",
)
@click.option("--wait-for-completion", type=bool, is_flag=True, default=False)
@click.option(
    "--local",
    is_flag=True,
    type=bool,
    default=False,
    help="If set, SageMaker LocalSession will be used to run the pipeline",
)
@click.option(
    "--auto-build",
    type=bool,
    is_flag=True,
    default=False,
    help="Specify to docker build and push before scheduling a run",
)
@click.option(
    "--yes",
    "-y",
    type=bool,
    is_flag=True,
    default=False,
    help="Auto answer yes confirm prompts",
)
@click.pass_obj
@click.pass_context
def run(
    click_context: click.Context,
    ctx: CliContext,
    execution_role: Optional[str],
    image: Optional[str],
    pipeline: str,
    params: str,
    wait_for_completion: bool,
    local: bool,
    auto_build: bool,
    yes: bool,
):
    """
    Runs the pipeline on SageMaker Pipelines
    """
    mgr: KedroContextManager
    with get_context_and_pipeline(
        ctx, image, pipeline, params, local, execution_role
    ) as (
        mgr,
        sm_pipeline,
    ):
        docker_autobuild(
            auto_build, click_context, image or mgr.plugin_config.docker.image, mgr, yes
        )

        client = SageMakerClient(
            sm_pipeline, execution_role or mgr.plugin_config.aws.execution_role
        )

        is_ok = client.run(
            is_local=local,
            wait_for_completion=wait_for_completion,
            on_pipeline_started=lambda p: click.echo(
                f"Pipeline ARN: {p.describe()['PipelineArn']}"
            ),
        )

        if is_ok:
            exit_code = 0
            click.echo(
                click.style(
                    "Pipeline {} successfully".format(
                        "finished" if wait_for_completion else "started"
                    ),
                    fg="green",
                )
            )
        else:
            exit_code = 1
            click.echo(
                click.style("There was an error while running the pipeline", fg="red")
            )

        click_context.exit(exit_code)


@sagemaker_group.command()
@click.option(
    "-r",
    "--execution-role",
    type=str,
    help="SageMaker execution role",
)
@click.option(
    "-i",
    "--image",
    type=str,
    help="Docker image to use for pipeline execution.",
)
@click.option(
    "-p",
    "--pipeline",
    "pipeline",
    type=str,
    help="Name of pipeline to run",
    default="__default__",
)
@click.option(
    "--params",
    "params",
    type=str,
    help="Parameters override in form of JSON string",
)
@click.option(
    "--local",
    is_flag=True,
    type=bool,
    default=False,
    help="If set, SageMaker LocalSession will be used to run the pipeline",
)
@click.option(
    "--output-path",
    type=click.types.Path(exists=False, dir_okay=False),
    default="pipeline.json",
    help="Path to output pipeline JSON file. Defaults to ./pipeline.json",
)
@click.pass_obj
def compile(
    ctx: CliContext,
    execution_role: Optional[str],
    image: Optional[str],
    pipeline: str,
    params: str,
    local: bool,
    output_path: str,
):
    """
    Compiles the pipeline to a JSON file
    """
    with get_context_and_pipeline(
        ctx, image, pipeline, params, local, execution_role
    ) as (
        _,
        sm_pipeline,
    ):
        target_path = Path(output_path)
        with target_path.open("w") as f:
            json.dump(json.loads(sm_pipeline.definition()), f, indent=4)
        click.echo(f"Pipeline compiled to {target_path}")


@sagemaker_group.command(
    hidden=True,
    context_settings=dict(ignore_unknown_options=True, allow_extra_args=True),
)
@click.pass_obj
@click.pass_context
def entrypoint(click_context: click.Context, ctx: CliContext, *other_args, **kwargs):
    """
    Internal entrypoint only for use with Kedro SageMaker plugin
    """
    is_debug = int(os.environ.get(KEDRO_SAGEMAKER_DEBUG, "0"))
    if is_debug:
        print("\n".join(sys.argv[1:]))
        print("-" * 80)
        print(json.dumps(dict(os.environ), indent=4))

    args = os.environ.get(KEDRO_SAGEMAKER_ARGS)

    os.chdir(os.environ.get(KEDRO_SAGEMAKER_WORKING_DIRECTORY))
    kedro_params = parse_flat_parameters(os.environ)

    result = subprocess.run(
        shlex.join(shlex.split(args) + ["--params", json.dumps(kedro_params)]),
        stdout=sys.stdout,
        stderr=subprocess.STDOUT,
        shell=True,
        check=False,
    )

    click_context.exit(result.returncode)


@sagemaker_group.command(hidden=True)
@click.option(
    "-p",
    "--pipeline",
    "pipeline",
    type=str,
    help="Name of pipeline to run",
    default="__default__",
)
@click.option(
    "-n", "--node", "node", type=str, help="Name of the node to run", required=True
)
@click.option(
    "--params",
    "params",
    type=str,
    help="Parameters override in form of JSON string",
)
@click.pass_obj
def execute(ctx: CliContext, pipeline: str, node: str, params: str):
    parameters = parse_extra_params(params)
    with KedroContextManager(
        ctx.metadata.package_name, env=ctx.env, extra_params=parameters
    ) as mgr:
        if is_mlflow_enabled():
            env_key, env_value = lookup_mlflow_run_id(
                mgr.context, os.getenv(KEDRO_SAGEMAKER_EXECUTION_ARN)
            )
            if env_value is not None:
                click.echo(f"Mlflow run id: {env_value}")
                os.environ[env_key] = env_value

        runner = SageMakerPipelinesRunner()
        mgr.session.run(pipeline, node_names=[node], runner=runner)


@sagemaker_group.command(hidden=True)
@click.pass_obj
def mlflow_start(ctx: CliContext):
    """
    Registers new mlflow run with Sagemaker Execution ARN inside the tags
    """
    import mlflow
    from kedro_mlflow.config.kedro_mlflow_config import KedroMlflowConfig

    with KedroContextManager(ctx.metadata.package_name, env=ctx.env) as mgr:
        mlflow_conf: KedroMlflowConfig = mgr.context.mlflow

    run = mlflow.start_run(
        experiment_id=mlflow.get_experiment_by_name(
            mlflow_conf.tracking.experiment.name
        ).experiment_id,
        nested=False,
    )
    mlflow.set_tag(MLFLOW_TAG_EXECUTION_ARN, os.environ[KEDRO_SAGEMAKER_EXECUTION_ARN])
    click.echo(f"Started run: {run.info.run_id}")



import logging
from typing import Callable, Optional

from sagemaker.workflow.pipeline import Pipeline as SageMakerPipeline

logger = logging.getLogger(__name__)


class SageMakerClient:
    def __init__(self, sagemaker_pipeline: SageMakerPipeline, execution_role: str):
        self.execution_role = execution_role
        self.sagemaker_pipeline = sagemaker_pipeline

    def run(
        self,
        is_local: bool,
        wait_for_completion: bool = False,
        on_pipeline_started: Optional[Callable[[SageMakerPipeline], None]] = None,
    ):
        self.sagemaker_pipeline.upsert(self.execution_role)
        smp = self.sagemaker_pipeline.start()
        if on_pipeline_started:
            on_pipeline_started(smp)

        if not is_local and wait_for_completion:
            try:
                smp.wait()
                return True
            except Exception:
                logger.exception("Error while running the pipeline", exc_info=True)
                return False
        else:
            return True

    def update(self):
        self.sagemaker_pipeline.upsert(self.execution_role)
